import traverse from '@babel/traverse';
import * as t from '@babel/types';
import * as RA from 'effect/Array';
import * as Effect from 'effect/Effect';
import { pipe } from 'effect/Function';
import * as Hash from 'effect/Hash';
import * as Option from 'effect/Option';
import { applyParentEntries } from '@native-twin/css/jsx';
import { Tree } from '@native-twin/helpers/tree';
import { mappedComponents, createCommonMappedAttribute } from '../../utils/component.maps';
import * as jsxPredicates from './jsx.predicates';

const getJSXElementSource = (path) => pipe(getJSXElementName(path.node.openingElement), Option.flatMap((x) => Option.fromNullable(path.scope.getBinding(x))), Option.flatMap((binding) => getBindingImportSource(binding)), Option.getOrElse(() => ({ kind: 'local', source: 'unknown' })));
const getBindingImportDeclaration = (binding) => pipe(binding.path, Option.liftPredicate(jsxPredicates.isImportSpecifier), Option.bindTo('importSpecifier'), Option.bind('importDeclaration', ({ importSpecifier }) => Option.liftPredicate(importSpecifier.parentPath, jsxPredicates.isImportDeclaration)), Option.map((source) => ({
    kind: 'import',
    source: source.importDeclaration.node.source.value,
})));
const getBindingRequireDeclaration = (binding) => pipe(binding.path, Option.liftPredicate(jsxPredicates.isVariableDeclaratorPath), Option.bindTo('importSpecifier'), Option.bind('requireExpression', ({ importSpecifier }) => pipe(Option.fromNullable(importSpecifier.node.init), Option.flatMap((init) => Option.liftPredicate(init, jsxPredicates.isCallExpression)), Option.flatMap((x) => RA.head(x.arguments)), Option.flatMap((x) => Option.liftPredicate(x, t.isStringLiteral)))), Option.map((source) => {
    return {
        kind: 'require',
        source: source.requireExpression.value,
    };
}));
const getBindingImportSource = (binding) => pipe([getBindingImportDeclaration(binding), getBindingRequireDeclaration(binding)], Option.firstSomeOf);
/**
 * @internal
 * @category Transformer
 * @domain Babel
 * Extract {@link JSXMappedAttribute[]} list from a jsx Attribute
 * */
const getJSXMappedAttributes = (attributes, config) => {
    return attributes
        .map((x) => extractStyledProp(x, config))
        .filter((x) => x !== null);
};
/**
 * @category Transformer
 * @domain Babel
 * Extract {@link JSXMappedAttribute[]} list from a jsx Attribute
 * */
const extractMappedAttributes = (node) => {
    const attributes = getJSXElementAttrs(node);
    return pipe(getJSXElementName(node.openingElement), Option.flatMap((x) => Option.fromNullable(getJSXElementConfig(x))), Option.map((mapped) => getJSXMappedAttributes(attributes, mapped)), Option.getOrElse(() => []));
};
/**
 * * @internal
 * @domain Babel
 * @description Extract the {@link MappedComponent} from any {@link ValidJSXElementNode}
 * */
const getJSXElementConfig = (tagName) => {
    const componentConfig = mappedComponents.find((x) => x.name === tagName);
    if (!componentConfig)
        return createCommonMappedAttribute(tagName);
    return componentConfig;
};
/**
 * @domain Babel
 * @description Extract the {@link t.JSXAttribute[]} from any {@link t.JSXElement}
 * */
const getJSXElementAttrs = (element) => pipe(element.openingElement.attributes, RA.filter(jsxPredicates.isJSXAttribute));
/**
 * @domain Babel
 * @description Extract the {@link JSXMappedAttribute} from any {@link t.JSXAttribute}
 * */
const extractStyledProp = (attribute, config) => {
    const validClassNames = Object.entries(config.config);
    if (!t.isJSXAttribute(attribute))
        return null;
    if (!t.isJSXIdentifier(attribute.name))
        return null;
    const className = validClassNames.find((x) => attribute.name.name === x[0]);
    if (!className)
        return null;
    if (t.isStringLiteral(attribute.value)) {
        return {
            prop: className[0],
            target: className[1],
            value: attribute.value,
        };
    }
    if (t.isJSXExpressionContainer(attribute.value)) {
        if (t.isTemplateLiteral(attribute.value.expression)) {
            return {
                prop: className[0],
                target: className[1],
                value: attribute.value.expression,
            };
        }
        if (t.isCallExpression(attribute.value.expression)) {
            return {
                prop: className[0],
                target: className[1],
                value: t.templateLiteral([
                    t.templateElement({ raw: '', cooked: '' }),
                    t.templateElement({ raw: '', cooked: '' }),
                ], [attribute.value.expression]),
            };
        }
    }
    return null;
};
const getJSXElementName = (openingElement) => {
    if (t.isJSXIdentifier(openingElement.name)) {
        return Option.some(openingElement.name.name);
    }
    return Option.none();
};
const getBabelJSXElementChildsCount = (node) => pipe(node.children, RA.filterMap((x) => pipe(x, Option.liftPredicate(t.isJSXElement))), RA.length);
const getAstTrees = (ast, filename) => {
    return Effect.async((resolve) => {
        const cssImports = [];
        traverse(ast, {
            Program: {
                exit() {
                    resolve(Effect.succeed(this.trees));
                },
            },
            ImportDeclaration(path) {
                if (path.node.source.value.endsWith('.css')) {
                    cssImports.push(path.node.source.value);
                }
            },
            JSXElement(path) {
                const uid = path.scope.generateUidIdentifier();
                // console.log('UID: ', uid.name, uidUnNamed.name);
                const hash = Hash.string(filename + uid.name);
                const parentTree = new Tree({
                    order: -1,
                    babelNode: path.node,
                    uid: `${hash}__${uid.name}:__1`,
                    source: getJSXElementSource(path),
                    cssImports: cssImports,
                    parentID: null,
                });
                gelBabelJSXElementChildLeaves(path, parentTree.root);
                this.trees.push(parentTree);
                path.skip();
            },
        }, undefined, {
            trees: [],
        });
    });
};
const gelBabelJSXElementChildLeaves = (path, parent) => {
    const childs = pipe(path.get('children'), RA.filterMap(Option.liftPredicate(jsxPredicates.isJSXElementPath)));
    for (const childPath of childs) {
        const order = parent.childrenCount;
        const childUid = path.scope.generateUid();
        // console.log('CHILD_UID: ', childUid);
        const childLeave = parent.addChild({
            order,
            babelNode: childPath.node,
            uid: `${parent.value.uid}:${order}____${childUid}`,
            source: getJSXElementSource(childPath),
            cssImports: parent.value.cssImports,
            parentID: parent.value.uid,
        });
        childLeave.parent = parent;
        gelBabelJSXElementChildLeaves(childPath, childLeave);
    }
};
const getJSXCompiledTreeRuntime = (leave, parentLeave) => {
    const runtimeSheet = pipe(parentLeave, Option.map((parent) => applyParentEntries(leave.entries, parent.childEntries, leave.order, leave.parentSize)), Option.getOrElse(() => leave.entries));
    return {
        leave,
        runtimeSheet,
    };
};

export { extractMappedAttributes, extractStyledProp, gelBabelJSXElementChildLeaves, getAstTrees, getBabelJSXElementChildsCount, getBindingImportSource, getJSXCompiledTreeRuntime, getJSXElementAttrs, getJSXElementName, getJSXElementSource };
//# sourceMappingURL=jsx.maps.js.map
