import * as RA from 'effect/Array';
import { pipe, apply, flip } from 'effect/Function';
import * as Predicate from 'effect/Predicate';

function expressionFactory(writer) {
    const curried = {
        object: curryWriter(objectWriter),
        objectProperty: curryWriter(createKeyValuePair),
        primitive: curryWriter(createPrimitive),
        identifier: curryWriter(createIdentifier),
        nextToken: curryWriter(createNextToken),
        array: curryWriter(arrayWriter),
        writer,
    };
    return curried;
    function curryWriter(cb) {
        return pipe(cb, apply(writer));
    }
    function createNextToken(writer) {
        return (x) => {
            if (isStringLiteral(x)) {
                return curried.identifier(x);
            }
            if (isPrimitive(x)) {
                return curried.primitive(x);
            }
            if (RA.isArray(x))
                return curried.array(x);
            if (Predicate.isRecord(x))
                return curried.object(x);
            if (Predicate.isNullable(x))
                return curried.identifier(`${x}`);
            return writer;
        };
    }
    function createIdentifier(writer) {
        return (x) => {
            return writer.write(`${x}`);
        };
    }
    function createPrimitive(writer) {
        return (primitive) => {
            if (typeof primitive === 'string') {
                return writer.quote(primitive);
            }
            return writer.write(`${primitive}`);
        };
    }
    function createKeyValuePair(writer) {
        return ([key, value]) => {
            return writer.hangingIndent(() => pipe(writer.write(key).write(':').space(), flip(createNextToken)(value)));
        };
    }
    function arrayWriter(writer) {
        return (arrayValue) => {
            return pipe(writer.write('['), (w) => pipe(arrayValue, RA.forEach((ww, i) => {
                curried.nextToken(ww).conditionalWrite(i + 1 < arrayValue.length, ',');
            }), () => w), (w) => w.write(']'));
        };
    }
    function objectWriter(writer) {
        return (object) => {
            const entries = pipe(object, Object.entries, RA.filter((x) => !Predicate.isFunction(x[1]) && !Predicate.isPromise(x[1])));
            return writer.block(() => {
                pipe(entries, RA.forEach((x, i) => {
                    pipe(x, curried.objectProperty, (w) => w
                        .conditionalWrite(i + 1 < entries.length, ',')
                        .conditionalNewLine(i < entries.length));
                }));
            });
        };
    }
}
const isStringLiteralLike = (s) => s.startsWith('`');
const isStringLiteral = pipe(Predicate.isString, Predicate.compose(isStringLiteralLike));
const isStringOrNumber = Predicate.or(Predicate.isString, Predicate.isNumber);
const isPrimitive = Predicate.or(isStringOrNumber, Predicate.isBoolean);

export { expressionFactory };
//# sourceMappingURL=writer.factory.js.map
